print("Importing Standard Libraries...", end=" ")
import cv2
import numpy as np
import math
import serial
import argparse
print("[DONE]")

# Import and initialize Camera
from lib.camera import Camera
cam = Camera()

# Initialize serial output
ser = serial.Serial('/dev/ttyS0')
ser.reset_input_buffer()
ser.reset_output_buffer()

# add argument parser
parser = argparse.ArgumentParser()
parser.add_argument('-nw', '--no_window', action='store_true') # argument to supress window
args = parser.parse_args()

'''
Detect the shape of the found contour
'''
def detectShape(cnt):
    '''
    1. Shape: Circle
    Enclose the contours with the smallest possible circle.
    If the area of this circle is approximately the same as the area of the contour it is a circle
    '''    
    (x,y),radius = cv2.minEnclosingCircle(cnt)
    bounding_area = radius**2 * math.pi
    detected_area = cv2.contourArea(cnt) + 0.000001 # add a very small number to avoid division by zero
    area_ratio = bounding_area/detected_area

    if area_ratio >= 0.80 and area_ratio <= 1.2: 
        return 'circle'
    
    '''
    2. Shape: Rectangle
    Enclose the contours with the smallest possible rectangle.
    The rectangle is not a square and has a aspect ratio of approximately 0.5
    '''
    x, y, width, height = cv2.boundingRect(cnt)
    aspect_ratio = height / float(width)
    if aspect_ratio >= 0.4 and aspect_ratio <= 0.6:
        return 'rectangle'
    
    '''
    3. Shape: Triangle
    The triangle is equilateral. That means, that the ratio between the triangle and the smallest enclosing circle
    is 4*Pi/(3*sqrt(3)). So we use the same area ratio from the circle calculation above and compare it to this constant.
    '''    
    triangle_ratio = 4*math.pi/(3*math.sqrt(3))        
    if area_ratio >= 0.8*triangle_ratio and area_ratio <= 1.2*triangle_ratio:
        return 'triangle'          

    # if no shape was found, return 'None'
    return 'None'


'''
Endless loop 
'''
print("Start Shape Detection Process")
while True:
    print("NEW IMAGE", end=' ')
    image = cam.read() # read an image
    
    # convert the color image into grayscale
    grayScale = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    
    # Binarize with Difference of Gaussians (DoG)
    gw, gs, gw1, gs1, gw2, gs2 = (3, 1.0, 7, 3.0, 3, 2.0)
    img_blur = cv2.GaussianBlur(grayScale, (gw,gw), gs)
    g1 = cv2.GaussianBlur(img_blur, (gw1,gw1), gs1)
    g2 = cv2.GaussianBlur(img_blur, (gw2,gw2), gs2)
    ret, th = cv2.threshold(g2-g1,127,255,cv2.THRESH_BINARY)
    

    # Find edges in the image using canny edge detection method
    # Calculate lower threshold and upper threshold using sigma = 0.33
    sigma = 0.03
    v = np.median(th)
    low = int(max(0, (1.0 - sigma) * v))
    high = int(min(255, (1.0 + sigma) * v))
    
    edged = cv2.Canny(th, low, high)

    # After finding edges we have to find contours
    # Contour is a curve of points with no gaps in the curve
    # It will help us to find location of shapes
    
    # cv2.RETR_EXTERNAL is passed to find the outermost contours (because we want to outline the shapes)
    # cv2.CHAIN_APPROX_SIMPLE is removing redundant points along a line
    (_, cnts, _) = cv2.findContours(edged, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    cnts = [x for x in cnts if cv2.contourArea(x) > 80] # throw away all contours that are smaller than the specified value
    
    # loop through all contours
    for c in cnts:       
        # call detectShape for contour c
        shape = detectShape(c)
        
        # if a shape is found send the name of the shape over the serial communication
        if shape != 'None':
            if not args.no_window:
                # Outline the contours
                cv2.drawContours(image, [c], -1, (0, 255, 0), 1)                
            print(shape)
            ser.write(shape.encode())
            ser.write(b'\n')
    print('')
    
    # show a live feed of the camera
    if not args.no_window:
        cv2.imshow("frame", cv2.resize(image, (640,480)))
        # quit the live feed with 'q'
        if chr(cv2.waitKey(1)&255) == 'q':
            cv2.destroyAllWindows()
            cam.close()
            break
